import numpy as np
from functools import wraps


try:
    # fast versions
    import bottleneck as bn

    def _wrap_function(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            out = kwargs.pop('out', None)
            data = f(*args, **kwargs)
            if out is None:
                out = data
            else:
                out[()] = data

            return out

        return wrapped

    nanmean = _wrap_function(bn.nanmean)
    nanstd = _wrap_function(bn.nanstd)
    nansum = _wrap_function(bn.nansum)
    nanmax = _wrap_function(bn.nanmax)
    nanmin = _wrap_function(bn.nanmin)
    nanargmax = _wrap_function(bn.nanargmax)
    nanargmin = _wrap_function(bn.nanargmin)
except ImportError:
    # slower numpy
    nanmean = np.nanmean
    nanstd = np.nanstd
    nansum = np.nansum
    nanmax = np.nanmax
    nanmin = np.nanmin
    nanargmax = np.nanargmax
    nanargmin = np.nanargmin
    
    
def rolling_window(array, length, mutable=False):
    """
    Restride an array of shape

        (X_0, ... X_N)

    into an array of shape

        (length, X_0 - length + 1, ... X_N)

    where each slice at index i along the first axis is equivalent to

        result[i] = array[length * i:length * (i + 1)]

    Parameters
    ----------
    array : np.ndarray
        The base array.
    length : int
        Length of the synthetic first axis to generate.
    mutable : bool, optional
        Return a mutable array? The returned array shares the same memory as
        the input array. This means that writes into the returned array affect
        ``array``. The returned array also uses strides to map the same values
        to multiple indices. Writes to a single index may appear to change many
        values in the returned array.

    Returns
    -------
    out : np.ndarray

    Example
    -------
    >>> from numpy import arange
    >>> a = arange(25).reshape(5, 5)
    >>> a
    array([[ 0,  1,  2,  3,  4],
           [ 5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14],
           [15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24]])

    >>> rolling_window(a, 2)
    array([[[ 0,  1,  2,  3,  4],
            [ 5,  6,  7,  8,  9]],
    <BLANKLINE>
           [[ 5,  6,  7,  8,  9],
            [10, 11, 12, 13, 14]],
    <BLANKLINE>
           [[10, 11, 12, 13, 14],
            [15, 16, 17, 18, 19]],
    <BLANKLINE>
           [[15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24]]])
    """
    if not length:
        raise ValueError("Can't have 0-length window")

    orig_shape = array.shape
    if not orig_shape:
        raise IndexError("Can't restride a scalar.")
    elif orig_shape[0] < length:
        raise IndexError(
            "Can't restride array of shape {shape} with"
            " a window length of {len}".format(
                shape=orig_shape,
                len=length,
            )
        )

    num_windows = (orig_shape[0] - length + 1)
    new_shape = (num_windows, length) + orig_shape[1:]

    new_strides = (array.strides[0],) + array.strides

    out = as_strided(array, new_shape, new_strides)
    out.setflags(write=mutable)
    return out


# -*- coding: utf-8 -*-

from ultron.strategy.fut_strategy import Strategy as FutStrategy
from ultron.strategy.stk_strategy import Strategy as StkStrategy

func_dict = {'fut':FutStrategy,'stk':StkStrategy}
class StrategyEngine(object):
    @classmethod
    def create_class(cls, name):
        return func_dict[name] if name in func_dict else None


# -*- coding: utf-8 -*-

from ultron.tradingday.Date cimport Date
from ultron.tradingday.Period cimport Period


cdef class CalendarImpl(object):

    cdef bint isBizDay(self, Date date)
    cdef bint isWeekEnd(self, int weekDay)


cdef class Calendar(object):

    cdef public CalendarImpl _impl
    cdef public str name

    cpdef isBizDay(self, Date d)
    cpdef isHoliday(self, Date d)
    cpdef isWeekEnd(self, int weekday)
    cpdef isEndOfMonth(self, Date d)
    cpdef endOfMonth(self, Date d)
    cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=*, bint includeLast=*)
    cpdef adjustDate(self, Date d, int c=*)
    cpdef advanceDate(self, Date d, Period period, int c=*, bint endOfMonth=*)
    cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=*)
    cpdef bizDatesList(self, Date fromDate, Date toDate)


# -*- coding: utf-8 -*-

import numpy as np
from ....utilities.singleton import Singleton
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityCurrentValueHolder

##横截面变异
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcosValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcoshValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityTanhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySimpleReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogReturnValueHolder

##横截面交叉
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAddedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecuritySubbedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityMultipliedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDividedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityEqOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityNeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAndOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityOrOperatorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMinimumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMaximumValueHolder

from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder

from ultron.sentry.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder

## 横截面变异
#from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityPowValueHolder # 指定默认参数

## 时间序列变异
#from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMACDValueHolder #双默认参数

from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityXAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingDecay
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRank
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingQuantile
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAllTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAnyTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSum
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingVariance
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingStandardDeviation
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedPositive
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRSI
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingLogReturn

from ultron.sentry.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityShiftedValueHolder

## 时间序列交叉
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCorrelation

import six
from enum import Enum, unique


@unique
class FunctionType(Enum):
    cross_section = 1
    time_series = 2


class Function(object):

    def __init__(self, function, arity, ftype, default_value=0):
        self.function = function
        self.arity = arity
        self.name = function.__name__
        self.ftype = ftype
        self.default_value = default_value


@six.add_metaclass(Singleton)
class Operators(object):

    def __init__(self):

        # 时间序列默认周期列表
        self._ts_period = [2, 4, 6, 8, 10]

        self._init_cs()
        self._init_ts()

        self._cs_mutation_function_list = [
            Function(f, 1, FunctionType.cross_section)
            for f in self._cross_section_mutation_list
        ]
        self._cs_crossover_function_list = [
            Function(f, 2, FunctionType.cross_section)
            for f in self._cross_section_crossover_list
        ]
        #支持全部参数
        self._ts_mutation_function_list = []
        self._ts_crossover_function_list = []
        for period in self._ts_period:
            self._ts_mutation_function_list += [
                Function(f, 1, FunctionType.time_series, period)
                for f in self._time_series_mutation_list
            ]
            self._ts_crossover_function_list += [
                Function(f, 2, FunctionType.time_series, period)
                for f in self._time_series_crossover_list
            ]

        self._mutation_sets = self._cs_mutation_function_list + self._ts_mutation_function_list
        self._crossover_sets = self._cs_crossover_function_list + self._ts_crossover_function_list

    def _init_cs(self):
        self._cross_section_mutation_list = [
            SecurityAverageValueHolder, SecurityDiffValueHolder,
            SecurityTanhValueHolder, SecuritySignValueHolder,
            SecurityExpValueHolder, SecurityLogValueHolder,
            SecuritySqrtValueHolder, SecurityAbsValueHolder,
            SecurityAcosValueHolder, SecurityAsinValueHolder,
            SecurityAsinhValueHolder, SecurityAcoshValueHolder,
            SecurityNormInvValueHolder, SecurityCeilValueHolder,
            SecurityFloorValueHolder, SecurityRoundValueHolder,
            SecuritySimpleReturnValueHolder, SecurityLogReturnValueHolder,
            SecuritySigmoidValueHolder, CSRankedSecurityValueHolder,
            CSZScoreSecurityValueHolder, CSPercentileSecurityValueHolder,
            CSAverageSecurityValueHolder, CSAverageAdjustedSecurityValueHolder
        ]

        self._cross_section_crossover_list = [
            CSResidueSecurityValueHolder, SecurityAddedValueHolder,
            SecuritySubbedValueHolder, SecurityMultipliedValueHolder,
            SecurityDividedValueHolder, SecurityLtOperatorValueHolder,
            SecurityLeOperatorValueHolder, SecurityGtOperatorValueHolder,
            SecurityGeOperatorValueHolder, SecurityEqOperatorValueHolder,
            SecurityNeOperatorValueHolder, SecurityAndOperatorValueHolder,
            SecurityOrOperatorValueHolder, SecurityMinimumValueHolder,
            SecurityMaximumValueHolder
        ]

    def _init_ts(self):
        self._time_series_mutation_list = [
            SecurityXAverageValueHolder, SecurityMovingAverage,
            SecurityMovingDecay, SecurityMovingMax, SecurityMovingArgMax,
            SecurityMovingMin, SecurityMovingArgMin, SecurityMovingRank,
            SecurityMovingQuantile, SecurityMovingAllTrue,
            SecurityMovingAnyTrue, SecurityMovingSum, SecurityMovingVariance,
            SecurityMovingStandardDeviation, SecurityMovingCountedPositive,
            SecurityMovingPositiveAverage,
            SecurityMovingPositiveDifferenceAverage,
            SecurityMovingNegativeDifferenceAverage, SecurityMovingRSI,
            SecurityMovingLogReturn, SecurityDeltaValueHolder,
            SecurityShiftedValueHolder
        ]

        self._time_series_crossover_list = [SecurityMovingCorrelation]

    def custom_transformer(self, formula_sets):
        operators_sets = self._crossover_sets + self._mutation_sets
        return [
            operator for operator in operators_sets
            if operator.name in formula_sets
        ]

    def calc_factor(self,
                    expression,
                    total_data,
                    indexs,
                    key,
                    name='transformed',
                    dropna=False):
        return eval(expression).transform(total_data.set_index(indexs),
                                          category_field=key,
                                          name=name,
                                          dropna=dropna)

    def expression_params(self, expression):
        return eval(expression)._dependency

    def fetch_mutation_sets(self):
        return self._mutation_sets

    def fetch_crossover_sets(self):
        return self._crossover_sets


mutation_sets = Operators().fetch_mutation_sets()
crossover_sets = Operators().fetch_crossover_sets()
operators_sets = mutation_sets + crossover_sets
calc_factor = Operators().calc_factor
custom_transformer = Operators().custom_transformer
expression_params = Operators().expression_params



# -*- coding: utf-8 -*-

from typing import Tuple
import pandas as pd
from ultron.strategy.executor.base import Base

class NaiveExecutor(Base):
    def __init__(self):
        super(NaiveExecutor, self).__init__()
    
    def execute(self, target_pos: pd.DataFrame) -> Tuple[float, pd.DataFrame]:
        if self.current_pos.empty:
            turn_over = target_pos.weight.abs().sum()
        else:
            turn_over = self.calc_turn_over(target_pos, self.current_pos)
        self.current_pos = target_pos.copy()
        return turn_over, target_pos



# -*- coding: utf-8 -*-

from enum import IntEnum


class Frequency(IntEnum):

    NoFrequency = -1
    Once = 0
    Annual = 1
    Semiannual = 2
    EveryFourthMonth = 3
    Quarterly = 4
    Bimonthly = 6
    Monthly = 12
    EveryFourthWeek = 13
    Biweekly = 26
    Weekly = 52
    Daily = 365
    OtherFrequency = 999
